Require Import Ideal.


(** Trivial test *)

Definition infLoop := bmodule {{
  bfunction "f" [st ~> [< True >] ] {
    Halt
  }
}}.

Theorem infLoopOk : moduleOk infLoop.
  structured; eauto.
Qed.


(** Identity function *)

Definition ident := bmodule {{
  bfunction "f" [st ~> st#R0 @@ (st' ~> [<st'#R1 = st#R1>])] {
    Goto R0
  }
}}.

Theorem identOk : moduleOk ident.
  structured; eauto.
Qed.


(** Addition function *)

Definition add := bmodule {{
  bfunction "f" [st ~> st#R0 @@ (st' ~> [<st'#R1 = st#R1 + st#R2>])] {
    R1 <- R1 + R2;;
    Goto R0
  }
}}.

Theorem addOk : moduleOk add.
  structured; eauto.
Qed.


(** Iterative factorial *)

Require Import Arith.

Definition ifact := bmodule {{
  bfunction "fact" [st ~> st#Rret @@ (st' ~> [<st'#R1 = fact st#R0>])] {
    R1 <- 1;;
    R2 <- R0;;
    [st ~> [< fact st#R0 = st#R1 * fact st#R2 >]
      /\ st#Rret @@ (st' ~> [<st'#R1 = fact st#R0>])]
    While (R2 != 0) {
      R1 <- R1 * R2;;
      R2 <- R2 - 1
    };;
    Goto Rret
  }
}}.

Require Import Omega Ring.
Hint Extern 2 (@eq nat _ _) => omega.

Lemma fact_step : forall x y z,
  z <> 0
  -> x = y * fact z
  -> x = y * z * fact (z - 1).
  intros; subst; destruct z; simpl; omega || rewrite <- minus_n_O; ring.
Qed.

Hint Immediate fact_step.

Theorem ifactOk : moduleOk ifact.
  structured; eauto.
Qed.
