Require Import Ideal.


(** Swapping memory values *)

Definition swap := bmodule {{
  bfunction "swap" [st ~> Ex fr, Ex a, Ex b, ![ st#R0 ==> a * st#R1 ==> b * ![fr] ] st
    /\ st#Rret @@ (st' ~> ![ st#R1 ==> a * st#R0 ==> b * ![fr] ] st') ] {
    R2 <- $[R0];;
    $[R0] <- $[R1];;
    $[R1] <- R2;;
    Goto Rret
  }
}}.

Theorem swapOk : moduleOk swap.
  structuredSep; eauto.
Qed.


(** Basic list example *)

Module Type LINKED_LIST.
  Parameter lseg : list nat -> nat -> nat -> sprop.
  Parameter llist : list nat -> nat -> sprop.

  Axiom llistDefFwd : forall ls hd, llist ls hd ===> !{lseg ls hd O}.
  Axiom llistDefBwd : forall ls hd, !{lseg ls hd O} ===> llist ls hd.

  Axiom lsegEmpty : forall hd ls, lseg ls hd 0 ===> !{lseg nil hd hd} * !{lseg ls hd 0}.
  Axiom lsegMapNil : forall (f : nat -> nat) hd ls, ls = nil -> !{lseg nil hd hd} ===> lseg (map f ls) hd hd.
  Axiom lseg0 : forall ls, lseg ls 0 0 ===> [< ls = nil >].
  Axiom lsegS : forall hd ls, hd <> 0 -> lseg ls hd 0 ===> Ex x, Ex ls', Ex p, [< ls = x :: ls' >] * hd ==> x * (hd+1) ==> p * !{lseg ls' p 0}.
  Axiom lsegPostNil : forall (f : nat -> nat) hd ls, !{lseg (map f ls) hd 0} ===> lseg (map f (ls ++ nil)) hd 0.
  Axiom lsegAppend : forall hd tl x ls,
    Ex tl', [< tl' <> 0 >] * !{lseg (map S ls) hd tl'} * tl' ==> (x + 1) * (tl'+1) ==> tl ===> lseg (map S (ls ++ x :: nil)) hd tl.
End LINKED_LIST.

Module LinkedList : LINKED_LIST.
  Open Scope Sep_scope.

  Fixpoint lseg (ls : list nat) (hd tl : nat) : sprop :=
    match ls with
      | nil => [< hd = tl >]
      | x :: ls' => Ex p, [< hd <> 0 >] * hd ==> x * (hd+1) ==> p * !{lseg ls' p tl}
    end.

  Definition llist (ls : list nat) (hd : nat) := lseg ls hd O.

  Theorem llistDefFwd : forall ls hd, llist ls hd ===> !{lseg ls hd O}.
    sepLemma.
  Qed.

  Theorem llistDefBwd : forall ls hd, !{lseg ls hd O} ===> llist ls hd.
    sepLemma.
  Qed.

  Theorem lsegEmpty : forall hd ls, lseg ls hd 0 ===> !{lseg nil hd hd} * !{lseg ls hd 0}.
    sepLemma.
  Qed.

  Theorem lsegMapNil : forall (f : nat -> nat) hd ls, ls = nil -> !{lseg nil hd hd} ===> lseg (map f ls) hd hd.
    intros; subst; sepLemma.
  Qed.

  Theorem lseg0 : forall ls, lseg ls 0 0 ===> [< ls = nil >].
    destruct ls; sepLemma.
  Qed.

  Theorem lsegS : forall hd ls, hd <> 0 -> lseg ls hd 0
    ===> Ex x, Ex ls', Ex p, [< ls = x :: ls' >] * hd ==> x * (hd+1) ==> p * !{lseg ls' p 0}.
    destruct ls; sepLemma.
  Qed.

  Theorem lsegPostNil : forall (f : nat -> nat) hd ls, !{lseg (map f ls) hd 0} ===> lseg (map f (ls ++ nil)) hd 0.
    intros; rewrite <- app_nil_end; sepLemma.
  Qed.

  Lemma lsegAppend' : forall x tl ls hd,
    Ex tl', [< tl' <> 0 >] * !{lseg (map S ls) hd tl'} * tl' ==> (x + 1) * (tl'+1) ==> tl ===> lseg (map S (ls ++ x :: nil)) hd tl.
    induction ls; sepLemma; auto.
  Qed.

  Theorem lsegAppend : forall hd tl x ls,
    Ex tl', [< tl' <> 0 >] * !{lseg (map S ls) hd tl'} * tl' ==> (x + 1) * (tl'+1) ==> tl ===> lseg (map S (ls ++ x :: nil)) hd tl.
    intros; apply lsegAppend'.
  Qed.
End LinkedList.

Import LinkedList.

Definition linkedList := bmodule {{
  bfunction "linc" [st ~> Ex fr, Ex ls, ![ !{llist ls st#R0} * ![fr] ] st
    /\ st#Rret @@ (st' ~> ![ !{llist (map S ls) st#R0} * ![fr] ] st') ] {

    Use st [lsegEmpty st#R0];;
    Use st [lsegMapNil S st#R0];;

    [st ~> Ex fr, Ex hd, Ex lsPre, Ex lsPost, ![ !{lseg (map S lsPre) hd st#R0} * !{llist lsPost st#R0} * ![fr] ] st
      /\ st#Rret @@ (st' ~> ![ !{llist (map S (lsPre ++ lsPost)) hd} * ![fr] ] st') ]
    While (R0 != 0) {
      Use st [lsegS st#R0];;
      Use [lsegAppend];;

      $[R0] <- $[R0] + 1;;
      R0 <- $[R0+1]
    };;

    Use [lseg0];;
    Use [lsegPostNil S];;

    Goto Rret
  }
}}.

Hint Immediate llistDefFwd : Forward.
Hint Immediate llistDefBwd : Backward.

Lemma list_middle : forall A (x : A) ls2 ls1,
  (ls1 ++ x :: nil) ++ ls2 = ls1 ++ x :: ls2.
  induction ls1; simpl; intuition congruence.
Qed.

Hint Rewrite list_middle : PostSep.

Theorem linkedListOk : moduleOk linkedList.
  structuredSep; eauto.
Qed.
